import { InstanceExtension, InstanceState, Options, PluginInstance, PluginLifecycle, PluginMethod } from './models';
/**
 * Gets an instance for the given HTML Element
 *
 * @public
 * @param target - The HTML Element or a selector or an object with target
 */
export declare function find(target: HTMLElement | string | Pick<Options, 'target'>): Instance | null;
/**
 * Creates a new SpriteSpin instance, or updates an existing one
 *
 * @public
 */
export declare function createOrUpdate(options: Options): Instance;
/**
 * Creates a new SpriteSpin instance
 *
 * @public
 */
export declare function create(options: Options): Instance;
/**
 * Updates an existing instance with given options
 *
 * @public
 */
export declare function update(options: Options): Instance;
/**
 * Destroys the SpriteSpin instance
 *
 * @remarks
 * - stops running animation
 * - unbinds all events
 * - resets element state
 *
 * @public
 */
export declare function destroy(target: HTMLElement | string | Pick<Options, 'target'>): void;
/**
 * Adds methods to the SpriteSpin prototype
 *
 * @public
 */
export declare function extend<T extends InstanceExtension>(extension: T): void;
/**
 *
 * @public
 */
export declare class Instance {
    /**
     * The element this instance is bound to
     */
    readonly target: HTMLElement;
    /**
     * The current state of the spritespin instance
     */
    readonly state: InstanceState;
    private events;
    /**
     * Constructs a new spritespin instance and runs
     *
     * @param options - the configuration options
     */
    constructor(options: Options);
    /**
     * Updates the configuration and reboots the instance
     *
     * @param options - the new instance options
     */
    update(options: Partial<Options>): this;
    /**
     * Dispatches the `onInit` lifecycle event
     */
    init(): this;
    /**
     * Dispatches the `onFrame` and `onDraw` lifecycle events
     */
    tick(): this;
    /**
     * Dispatches the `onDestroy` lifecycle event
     */
    destroy(): this;
    /**
     * Preloads images and runs the initial lifecycle. Creates plugins, loads images, dispatches events.
     *
     * @public
     */
    load(): this;
    /**
     * Adds a lifecycle event listener
     *
     * @param event - the lifecycle event name
     * @param cb - the handler method
     * @returns a function to remove the added event listener
     */
    addListener(event: keyof PluginLifecycle, cb: PluginMethod): () => void;
    /**
     * Removes a previously added event listener
     *
     * @param cb - the handler to remove
     * @returns
     */
    removeListener(cb: PluginMethod): this;
    /**
     * Dispatches a lifecycle event
     *
     * @param event - the lifecycle method name
     */
    dispatch(event: keyof PluginLifecycle): void;
    /**
     * Gets the instance of an active plugin by name
     *
     * @param name - The name of the plugin
     */
    getPlugin(name: string): PluginInstance;
}
//# sourceMappingURL=instance.d.ts.map